import { warning, mapObj, mapObjIndexed } from './util';
import { view, over, set, lensProp, compose } from './util';

function prepareActions(context, actions = {}, actionGenerators = {}) {
  return Object.assign({}, actions, mapObj(o => o(context), actionGenerators));
}

function makeDispatchFn(scope, actionFn) {
  return function (action, ...params) {
    const [first, ...rest] = action.split('.');

    if (rest.length >= 1) {
      if (scope.modules.hasOwnProperty(first)) {
        const mod = scope.modules[first];
        mod.dispatch(rest.join('.'), ...params);
      } else {
        warning(false, `Unknown module on dispatch: ${scope.path || ''}.${first}`);
      }
    } else {
      if (scope.actions.hasOwnProperty(first)) {
        const actionPath = scope.path ? `${scope.path}.${action}` : action;
        const actionHandler = scope.actions[first];
        actionFn(scope, actionHandler, params, actionPath);
      } else {
        warning(false, `Unknown action: ${scope.path || ''}.${first}`);
      }
    }
  };
}

function makeAtomScope(parentScope, path, module, actionFn) {
  const subLens = lensProp(path);
  const lens = parentScope.lens ? compose(parentScope.lens, subLens) : subLens;
  const atom = parentScope.atom || parentScope;
  const subPath = parentScope.path ? `${parentScope.path}.${path}` : path;

  let scope = {
    atom,
    lens,
    path: subPath,
    swap: (fn, ...args) => atom.reset(over(lens, v => fn(v, ...args), atom.deref())),
    reset: newState => atom.reset(set(lens, newState, atom.deref())),
    deref: () => view(lens, atom.deref())
  };

  scope.modules = prepareModules(scope, module.modules);
  scope.actions = prepareActions(scope, module.actions, module.actionGenerators);
  scope.dispatch = makeDispatchFn(scope, actionFn);

  return scope;
}

function prepareModules(context, modules, actionFn) {
  return mapObjIndexed((module, name) => {
    return makeAtomScope(context, name, module, actionFn);
  }, modules);
}

function deriveInitialState(module) {
  return Object.assign({}, module.initialState || {}, recurseInitialState(module.modules || {}));
}

function recurseInitialState(modules) {
  return mapObj(deriveInitialState, modules);
}

export class Atom {
  constructor({
    initialState = {},
    actions = {},
    actionGenerators = {},
    modules = {}
  } = {}, middleware) {
    this.reset = newState => this._reset(this.deref(), newState);

    this.swap = (fn, ...args) => {
      const oldState = this.deref();
      this._reset(oldState, fn(oldState, ...args));
    };

    this.deref = () => this.state;

    this.addWatch = (watcherFn, limitToPaths = []) => {
      const exists = this.watches.indexOf(watcherFn) !== -1;
      warning(!exists, 'addWatch: watcher already watching');

      if (!exists) this.watches.push(watcherFn);
    };

    this.removeWatch = watcherFn => {
      const idx = this.watches.indexOf(watcherFn);
      const exists = idx !== -1;
      warning(exists, 'removeWatch: watcher not watching watching');

      if (exists) this.watches.splice(idx, 1);
    };

    this.notifyWatches = (oldState, newState) => {
      this.watches.forEach(sub => {
        // XXX: honour filtering
        sub(this, newState, oldState);
      });
    };

    this.atom = this;
    this.watches = [];
    initialState = Object.assign({}, initialState, recurseInitialState(modules));
    this.actions = prepareActions(this, actions, actionGenerators);

    if (typeof middleware === 'undefined') middleware = next => (...args) => next(...args);

    middleware((finalInitialState, resetFn, actionFn) => {
      this.modules = prepareModules(this, modules, actionFn);
      this.dispatch = makeDispatchFn(this, actionFn);
      this._reset = resetFn;
      this._reset({}, finalInitialState);
    })(initialState, (oldState, newState) => {
      this.state = newState;
      this.notifyWatches(oldState, newState);
    }, (scope, actionHandler, params, actionPath) => {
      if (typeof actionHandler === 'function') {
        actionHandler(scope, ...params);
      } else {
        warning(false, `Unknown action handler type for action: ${actionPath}`);
      }
    });
  }

}

// functional interface:
//
export function atom(...args) {
  let middleware;
  let module;

  if (typeof args[0] === 'function') {
    middleware = args[0];
    module = args[1];
  } else {
    module = args[0];
  }

  if (typeof module === 'undefined') {
    return function (module) {
      return new Atom(module, middleware);
    };
  } else {
    return new Atom(module, middleware);
  }
}

export function deref(atom) {
  return atom.deref();
}

export function swap(atom, fn, ...args) {
  atom.swap(fn, ...args);
}

export function reset(atom, newState) {
  atom.reset(newState);
}

export function dispatch(atom, action, ...args) {
  atom.dispatch(action, ...args);
}

export function addWatch(atom, watcherFn) {
  atom.addWatch(watcherFn);
}

export function removeWatch(atom, watcherFn) {
  atom.removeWatch(watcherFn);
}