'use strict';

exports.__esModule = true;
exports.Atom = undefined;
exports.atom = atom;
exports.deref = deref;
exports.swap = swap;
exports.reset = reset;
exports.dispatch = dispatch;
exports.addWatch = addWatch;
exports.removeWatch = removeWatch;

var _util = require('./util');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function prepareActions(context) {
  var actions = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var actionGenerators = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  return Object.assign({}, actions, (0, _util.mapObj)(function (o) {
    return o(context);
  }, actionGenerators));
}

function makeDispatchFn(scope, actionFn) {
  return function (action) {
    var _action$split = action.split('.'),
        first = _action$split[0],
        rest = _action$split.slice(1);

    for (var _len = arguments.length, params = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      params[_key - 1] = arguments[_key];
    }

    if (rest.length >= 1) {
      if (scope.modules.hasOwnProperty(first)) {
        var mod = scope.modules[first];
        mod.dispatch.apply(mod, [rest.join('.')].concat(params));
      } else {
        (0, _util.warning)(false, 'Unknown module on dispatch: ' + (scope.path || '') + '.' + first);
      }
    } else {
      if (scope.actions.hasOwnProperty(first)) {
        var actionPath = scope.path ? scope.path + '.' + action : action;
        var actionHandler = scope.actions[first];
        actionFn(scope, actionHandler, params, actionPath);
      } else {
        (0, _util.warning)(false, 'Unknown action: ' + (scope.path || '') + '.' + first);
      }
    }
  };
}

function makeAtomScope(parentScope, path, module, actionFn) {
  var subLens = (0, _util.lensProp)(path);
  var lens = parentScope.lens ? (0, _util.compose)(parentScope.lens, subLens) : subLens;
  var atom = parentScope.atom || parentScope;
  var subPath = parentScope.path ? parentScope.path + '.' + path : path;

  var scope = {
    atom: atom,
    lens: lens,
    path: subPath,
    swap: function swap(fn) {
      for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
        args[_key2 - 1] = arguments[_key2];
      }

      return atom.reset((0, _util.over)(lens, function (v) {
        return fn.apply(undefined, [v].concat(args));
      }, atom.deref()));
    },
    reset: function reset(newState) {
      return atom.reset((0, _util.set)(lens, newState, atom.deref()));
    },
    deref: function deref() {
      return (0, _util.view)(lens, atom.deref());
    }
  };

  scope.modules = prepareModules(scope, module.modules);
  scope.actions = prepareActions(scope, module.actions, module.actionGenerators);
  scope.dispatch = makeDispatchFn(scope, actionFn);

  return scope;
}

function prepareModules(context, modules, actionFn) {
  return (0, _util.mapObjIndexed)(function (module, name) {
    return makeAtomScope(context, name, module, actionFn);
  }, modules);
}

function deriveInitialState(module) {
  return Object.assign({}, module.initialState || {}, recurseInitialState(module.modules || {}));
}

function recurseInitialState(modules) {
  return (0, _util.mapObj)(deriveInitialState, modules);
}

var Atom = exports.Atom = function Atom() {
  var _this = this;

  var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
      _ref$initialState = _ref.initialState,
      initialState = _ref$initialState === undefined ? {} : _ref$initialState,
      _ref$actions = _ref.actions,
      actions = _ref$actions === undefined ? {} : _ref$actions,
      _ref$actionGenerators = _ref.actionGenerators,
      actionGenerators = _ref$actionGenerators === undefined ? {} : _ref$actionGenerators,
      _ref$modules = _ref.modules,
      modules = _ref$modules === undefined ? {} : _ref$modules;

  var middleware = arguments[1];

  _classCallCheck(this, Atom);

  this.reset = function (newState) {
    return _this._reset(_this.deref(), newState);
  };

  this.swap = function (fn) {
    for (var _len3 = arguments.length, args = Array(_len3 > 1 ? _len3 - 1 : 0), _key3 = 1; _key3 < _len3; _key3++) {
      args[_key3 - 1] = arguments[_key3];
    }

    var oldState = _this.deref();
    _this._reset(oldState, fn.apply(undefined, [oldState].concat(args)));
  };

  this.deref = function () {
    return _this.state;
  };

  this.addWatch = function (watcherFn) {
    var limitToPaths = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];

    var exists = _this.watches.indexOf(watcherFn) !== -1;
    (0, _util.warning)(!exists, 'addWatch: watcher already watching');

    if (!exists) _this.watches.push(watcherFn);
  };

  this.removeWatch = function (watcherFn) {
    var idx = _this.watches.indexOf(watcherFn);
    var exists = idx !== -1;
    (0, _util.warning)(exists, 'removeWatch: watcher not watching watching');

    if (exists) _this.watches.splice(idx, 1);
  };

  this.notifyWatches = function (oldState, newState) {
    _this.watches.forEach(function (sub) {
      // XXX: honour filtering
      sub(_this, newState, oldState);
    });
  };

  this.atom = this;
  this.watches = [];
  initialState = Object.assign({}, initialState, recurseInitialState(modules));
  this.actions = prepareActions(this, actions, actionGenerators);

  if (typeof middleware === 'undefined') middleware = function middleware(next) {
    return function () {
      return next.apply(undefined, arguments);
    };
  };

  middleware(function (finalInitialState, resetFn, actionFn) {
    _this.modules = prepareModules(_this, modules, actionFn);
    _this.dispatch = makeDispatchFn(_this, actionFn);
    _this._reset = resetFn;
    _this._reset({}, finalInitialState);
  })(initialState, function (oldState, newState) {
    _this.state = newState;
    _this.notifyWatches(oldState, newState);
  }, function (scope, actionHandler, params, actionPath) {
    if (typeof actionHandler === 'function') {
      actionHandler.apply(undefined, [scope].concat(params));
    } else {
      (0, _util.warning)(false, 'Unknown action handler type for action: ' + actionPath);
    }
  });
};

// functional interface:
//


function atom() {
  var middleware = void 0;
  var module = void 0;

  if (typeof (arguments.length <= 0 ? undefined : arguments[0]) === 'function') {
    middleware = arguments.length <= 0 ? undefined : arguments[0];
    module = arguments.length <= 1 ? undefined : arguments[1];
  } else {
    module = arguments.length <= 0 ? undefined : arguments[0];
  }

  if (typeof module === 'undefined') {
    return function (module) {
      return new Atom(module, middleware);
    };
  } else {
    return new Atom(module, middleware);
  }
}

function deref(atom) {
  return atom.deref();
}

function swap(atom, fn) {
  for (var _len4 = arguments.length, args = Array(_len4 > 2 ? _len4 - 2 : 0), _key4 = 2; _key4 < _len4; _key4++) {
    args[_key4 - 2] = arguments[_key4];
  }

  atom.swap.apply(atom, [fn].concat(args));
}

function reset(atom, newState) {
  atom.reset(newState);
}

function dispatch(atom, action) {
  for (var _len5 = arguments.length, args = Array(_len5 > 2 ? _len5 - 2 : 0), _key5 = 2; _key5 < _len5; _key5++) {
    args[_key5 - 2] = arguments[_key5];
  }

  atom.dispatch.apply(atom, [action].concat(args));
}

function addWatch(atom, watcherFn) {
  atom.addWatch(watcherFn);
}

function removeWatch(atom, watcherFn) {
  atom.removeWatch(watcherFn);
}