import React from 'react'
import PropTypes from 'prop-types'
import { Atom } from './atom'

export class Provider extends React.Component {
  static propTypes = {
    stateAtom: PropTypes.instanceOf(Atom),
    children: PropTypes.node.isRequired,
  }

  static childContextTypes = {
    stateAtom: PropTypes.instanceOf(Atom),
  }

  getChildContext() {
    return { stateAtom: this.stateAtom }
  }

  constructor(props, context) {
    super(props, context)
    this.stateAtom = props.stateAtom
  }

  render() {
    return React.Children.only(this.props.children)
  }
}

const getDisplayName = WrappedComponent =>
  WrappedComponent.displayName || WrappedComponent.name || 'Component'

export function connect(mapStateFn, mapActionsFn, equalFn = null) {
  if (typeof equalFn !== 'function') equalFn = (a, b) => a === b

  return function connectHOC(WrappedComponent) {
    return class extends React.PureComponent {
      static displayName = `BindAtom(${getDisplayName(WrappedComponent)})`

      static contextTypes = {
        stateAtom: PropTypes.instanceOf(Atom),
      }

      constructor(props, context) {
        super(props, context)

        if (mapStateFn)
          this.state = mapStateFn(context.stateAtom.deref(), props)
        if (mapActionsFn)
          this.actions = mapActionsFn(context.stateAtom, props)

        context.stateAtom.addWatch(this.stateUpdateHandler)
      }

      componentDidMount() {
        this.refreshActions(this.props)
        this.stateUpdateHandler()
      }

      componentWillReceiveProps(nextProps) {
        this.refreshActions(nextProps)
        if (mapStateFn)
          this.setState(mapStateFn(this.context.stateAtom.deref(), nextProps))
      }

      refreshActions(props) {
        if (mapActionsFn)
          this.actions = mapActionsFn(this.context.stateAtom, props)
      }

      componentWillUnmount() {
        this.context.stateAtom.removeWatch(this.stateUpdateHandler)
      }

      stateUpdateHandler = () => {
        if (mapStateFn) {
          const newState = mapStateFn(this.context.stateAtom.deref(), this.props)
          if (!equalFn(this.state, newState)) this.setState(newState)
        }
      }

      render() {
        return React.createElement(
          WrappedComponent,
          Object.assign({}, this.props, this.actions, this.state),
        )
      }
    }
  }
}

// const mapState = (state, ownProps) => ({
//  notes: state.notes,
// })
//
// const mapActions = (atom, ownProps) => ({
//   addNote: (...args) => atom.dispatch('notes.add', ...args),
// })
//
// export default connect(mapState, mapActions)(SomeComponent)
