import test from 'tape'
import sinon from 'sinon'
import { atom, swap, reset, deref, dispatch, addWatch, removeWatch, wrapGenerator, generatorMiddleware } from '../src/index'
import { mkAtomFixture1, mkAtomFixture2 } from './fixtures'

test('initial state', function(t) {
  const a = mkAtomFixture2(atom)

  t.deepEqual(deref(a), {a: "bcd", b: { val: 3} })

  t.end()
})

test('middleware can modify initial state', function(t) {
  const middleware = next => (initialState, resetFn, actionFn) => {
    const finalInitialState = Object.assign({}, initialState, { b: { val: 6 } })
    next(finalInitialState, resetFn, actionFn)
  }

  const enhancedAtom = atom(middleware)
  const a = mkAtomFixture2(enhancedAtom)

  t.deepEqual(deref(a), {a: "bcd", b: { val: 6} })

  t.end()
})

test('simple reset', function(t) {
  const a = atom({
    initialState: {
      a: "bcd",
    },
    actions: {},
  })

  reset(a, { a: "xyz" })

  t.deepEqual(deref(a), {a: "xyz"})

  t.end()
})


test('simple swap', function(t) {
  const a = atom({
    initialState: {
      a: "bcd",
    },
    actions: {},
  })

  swap(a, oldState => ({ ...oldState, a: oldState.a + "efg" }))

  t.deepEqual(deref(a), {
    a: "bcdefg",
  })

  t.end()
})


test('initial state with nested modules', function(t) {
  const cmod = {
    initialState: {
      cval: "ccc",
    },
    actions: {},
  }

  const bmod = {
    initialState: {
      bval: "bbb",
    },
    actions: {},
    modules: {
      c: cmod,
    }
  }

  const a = atom({
    initialState: {
      a: "bcd",
    },
    actions: {},
    modules: {
      b: bmod,
    }
  })

  t.deepEqual(deref(a), {
    a: "bcd",
    b: {
      bval: "bbb",
      c: {
        cval: "ccc",
      }
    }
  })

  t.end()
})

test('simple dispatch', function(t) {
  const a = mkAtomFixture1()

  dispatch(a, 'incBy', 9)

  t.deepEqual(deref(a), {
    a: 10,
    b: {
      val: 3,
      c: {}
    }
  })

  t.end()
})

test('simple dispatch via action generator and middleware', function(t) {
  const enhancedAtom = atom(generatorMiddleware)

  const a = enhancedAtom({
    initialState: {
      a: 1,
    },
    actions: {
      incBy: function(scope, incQty) {
        swap(scope, oldState => ({ ...oldState, a: oldState.a + incQty }))
      }
    },
    actionGenerators: {
      incByGen: wrapGenerator(function*(scope) {
        const [incQty] = yield
        swap(scope, oldState => ({ ...oldState, a: oldState.a + incQty }))
      })
    },
  })

  dispatch(a, 'incBy', 3)
  dispatch(a, 'incByGen', 6)

  t.deepEqual(deref(a), {
    a: 10,
  })

  t.end()
})

test('child dispatch', function(t) {
  const a = mkAtomFixture1()

  dispatch(a, 'b.decBy', 9)

  t.deepEqual(deref(a), {
    a: 1,
    b: {
      val: -6,
      c: {}
    }
  })

  t.end()
})

test('child dispatch with reset', function(t) {
  const bmod = {
    initialState: {
      val: 3,
    },
    actions: {
      doSomething: function(scope, newVal) {
        reset(scope, { val: newVal })
      }
    },
  }

  const a = atom({
    initialState: {
      a: 1,
    },
    modules: {
      b: bmod,
    },
    actions: {
      doSomething: function(scope, incQty) {
        swap(scope, oldState => ({ ...oldState, a: oldState.a + incQty }))
      }
    },
  })

  dispatch(a, 'b.doSomething', 9)

  t.deepEqual(deref(a), {
    a: 1,
    b: {
      val: 9,
    }
  })

  t.end()
})

test('simple watch', function(t) {
  const a = mkAtomFixture1()

  const h = function(scope, newState, oldState) {
    t.deepEqual(oldState, {
      a: 1,
      b: { val: 3, c: {} },
    })

    t.deepEqual(newState, {
      a: 1,
      b: { val: -6, c: {} },
    })
    t.end()
  }

  addWatch(a, h)

  dispatch(a, 'b.decBy', 9)

  removeWatch(a, h)
})
